use zerocopy::{AsBytes, FromBytes, FromZeroes, Unaligned};

use crate::{
    transaction, FilePointer, FileRange, RawFilePointer, ReaderTrait, TransactionHandle, U64,
};

#[derive(Clone, Copy, FromBytes, FromZeroes, AsBytes, Unaligned)]
#[repr(transparent)]
struct Str {
    data: FileRange,
}

impl Str {
    fn new() -> Self {
        Self {
            data: RawFilePointer::null().range(0),
        }
    }
}

impl Str {
    fn set<R>(self, transaction: &mut TransactionHandle<R>, s: &str) -> Self {
        let data = if s.is_empty() {
            Str::new().data
        } else {
            let (range, data) = transaction.allocate_range(s.len() as u64);
            data.copy_from_slice(s.as_bytes());
            range
        };

        if self.data.len() != 0 {
            transaction.free_range(self.data);
        }

        Str { data }
    }

    fn get(self, reader: &impl ReaderTrait) -> &str {
        std::str::from_utf8(reader.read_raw(self.data)).unwrap()
    }
}

impl FilePointer<Str> {
    fn set<R>(self, transaction: &mut TransactionHandle<R>, s: &str) -> FilePointer<Str> {
        let new_str = transaction.read::<Str>(self).set(transaction, s);
        let (ptr, data) = transaction.modify(self);
        *data = new_str;
        ptr
    }

    fn get(self, reader: &impl ReaderTrait) -> &str {
        reader.read(self).get(reader)
    }
}
